@@ -20,9 +20,6 @@
 extern int optind;
 extern char *optarg;
 #endif
-#ifdef HAVE_OPTRESET
-extern int optreset;		/* defined by BSD, but not others */
-#endif
 #ifdef HAVE_ERRNO_H
 #include <errno.h>
 #endif
@@ -101,10 +98,7 @@ void do_open_filesys(int argc, char **argv)
 	blk_t	blocksize = 0;
 	int open_flags = 0;
 	
-	optind = 0;
-#ifdef HAVE_OPTRESET
-	optreset = 1;		/* Makes BSD getopt happy */
-#endif
+	reset_getopt();
 	while ((c = getopt (argc, argv, "iwfcb:s:")) != EOF) {
 		switch (c) {
 		case 'i':
@@ -242,10 +236,7 @@ void do_show_super_stats(int argc, char *argv[])
 	int	numdirs = 0;
 	const char *usage = "Usage: show_super [-h]";
 
-	optind = 0;
-#ifdef HAVE_OPTRESET
-	optreset = 1;		/* Makes BSD getopt happy */
-#endif
+	reset_getopt();
 	while ((c = getopt (argc, argv, "h")) != EOF) {
 		switch (c) {
 		case 'h':
@@ -1463,6 +1454,8 @@ void do_bmap(int argc, char *argv[])
 		return;
 
 	ino = string_to_inode(argv[1]);
+	if (!ino)
+		return;
 	blk = parse_ulong(argv[2], argv[0], "logical_block", &err);
 
 	errcode = ext2fs_bmap(current_fs, ino, 0, 0, 0, blk, &pblk);
@@ -1474,6 +1467,38 @@ void do_bmap(int argc, char *argv[])
 	printf("%d\n", pblk);
 }
 
+void do_imap(int argc, char *argv[])
+{
+	ext2_ino_t	ino;
+	unsigned long 	group, block, block_nr, offset;
+
+	if (common_args_process(argc, argv, 2, 2, argv[0],
+				"<file>", 0))
+		return;
+	ino = string_to_inode(argv[1]);
+	if (!ino)
+		return;
+
+	group = (ino - 1) / EXT2_INODES_PER_GROUP(current_fs->super);
+	offset = ((ino - 1) % EXT2_INODES_PER_GROUP(current_fs->super)) *
+		EXT2_INODE_SIZE(current_fs->super);
+	block = offset >> EXT2_BLOCK_SIZE_BITS(current_fs->super);
+	if (!current_fs->group_desc[(unsigned)group].bg_inode_table) {
+		com_err(argv[0], 0, "Inode table for group %d is missing\n",
+			group);
+		return;
+	}
+	block_nr = current_fs->group_desc[(unsigned)group].bg_inode_table + 
+		block;
+	offset &= (EXT2_BLOCK_SIZE(current_fs->super) - 1);
+
+	printf("Inode %d is part of block group %lu\n"
+	       "\tlocated at block %lu, offset 0x%04lx\n", ino, group,
+	       block_nr, offset);
+
+}
+
+
 
 static int source_file(const char *cmd_file, int sci_idx)
 {
@@ -1576,6 +1601,7 @@ int main(int argc, char **argv)
 		ss_perror(sci_idx, retval, "creating invocation");
 		exit(1);
 	}
+	ss_get_readline(sci_idx);
 
 	(void) ss_add_request_table (sci_idx, &ss_std_requests, 1, &retval);
 	if (retval) {
