@@ -22,6 +22,7 @@
 extern ext2_filsys current_fs;
 extern ext2_ino_t	root, cwd;
 
+extern void reset_getopt(void);
 extern FILE *open_pager(void);
 extern void close_pager(FILE *stream);
 extern int check_fs_open(char *name);
@@ -111,4 +112,5 @@ extern void do_show_debugfs_params(int argc, char **argv);
 extern void do_expand_dir(int argc, char **argv);
 extern void do_features(int argc, char **argv);
 extern void do_bmap(int argc, char **argv);
+extern void do_imap(int argc, char **argv);
 
