@@ -26,9 +26,11 @@
 #if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
 #include "ext2fs.h"
+#include "blkid.h"
 #else
 #include "ext2fs/ext2_fs.h"
 #include "ext2fs/ext2fs.h"
+#include "blkid/blkid.h"
 #endif
 
 #ifdef ENABLE_NLS
@@ -40,13 +42,17 @@
 #else
 #define N_(a) (a)
 #endif
-/* FIXME */
+#define P_(singular, plural, n) (ngettext (singular, plural, n))
+#ifndef NLS_CAT_NAME
 #define NLS_CAT_NAME "e2fsprogs"
+#endif
+#ifndef LOCALEDIR
 #define LOCALEDIR "/usr/share/locale"
-/* FIXME */
+#endif
 #else
 #define _(a) (a)
 #define N_(a) a
+#define P_(singular, plural, n) ((n) == 1 ? (singular) : (plural))
 #endif
 
 /*
@@ -156,6 +162,7 @@ struct resource_track {
 #define E2F_FLAG_JOURNAL_INODE	0x0080 /* Create a new ext3 journal inode */
 #define E2F_FLAG_SB_SPECIFIED	0x0100 /* The superblock was explicitly 
 					* specified by the user */
+#define E2F_FLAG_RESTARTED	0x0100 /* E2fsck has been restarted */
 
 /*
  * Defines for indicating the e2fsck pass number
@@ -180,8 +187,8 @@ typedef struct e2fsck_struct *e2fsck_t;
 struct e2fsck_struct {
 	ext2_filsys fs;
 	const char *program_name;
-	const char *filesystem_name;
-	const char *device_name;
+	char *filesystem_name;
+	char *device_name;
 	int	flags;		/* E2fsck internal flags */
 	int	options;
 	blk_t	use_superblock;	/* sb requested by user */
@@ -189,6 +196,7 @@ struct e2fsck_struct {
 	int	blocksize;	/* blocksize */
 	blk_t	num_blocks;	/* Total number of blocks */
 	int	mount_flags;
+	blkid_cache blkid;	/* blkid cache */
 
 #ifdef HAVE_SETJMP_H
 	jmp_buf	abort_loc;
@@ -273,7 +281,7 @@ struct e2fsck_struct {
 	 * ext3 journal support
 	 */
 	io_channel	journal_io;
-	const char	*journal_name;
+	char	*journal_name;
 
 #ifdef RESOURCE_TRACK
 	/*
@@ -323,6 +331,11 @@ struct e2fsck_struct {
 typedef __u32 region_addr_t;
 typedef struct region_struct *region_t;
 
+#ifndef HAVE_STRNLEN
+#define strnlen(str, x) e2fsck_strnlen((str),(x))
+extern int e2fsck_strnlen(const char * s, int count);
+#endif
+
 /*
  * Procedure declarations
  */
@@ -400,6 +413,8 @@ extern int e2fsck_reconnect_file(e2fsck_t ctx, ext2_ino_t inode);
 extern errcode_t e2fsck_expand_directory(e2fsck_t ctx, ext2_ino_t dir,
 					 int num, int gauranteed_size);
 extern ext2_ino_t e2fsck_get_lost_and_found(e2fsck_t ctx, int fix);
+extern errcode_t e2fsck_adjust_inode_count(e2fsck_t ctx, ext2_ino_t ino, 
+					   int adj);
 
 
 /* region.c */
@@ -427,6 +442,7 @@ extern void fatal_error(e2fsck_t ctx, const char * fmt_string);
 extern void e2fsck_read_bitmaps(e2fsck_t ctx);
 extern void e2fsck_write_bitmaps(e2fsck_t ctx);
 extern void preenhalt(e2fsck_t ctx);
+extern char *string_copy(e2fsck_t ctx, const char *str, int len);
 #ifdef RESOURCE_TRACK
 extern void print_resource_track(const char *desc,
 				 struct resource_track *track);
@@ -446,3 +462,5 @@ extern int ext2_file_type(unsigned int mode);
 
 /* unix.c */
 extern void e2fsck_clear_progbar(e2fsck_t ctx);
+extern int e2fsck_simple_progress(e2fsck_t ctx, const char *label,
+				  float percent, unsigned int dpynum);
