@@ -38,7 +38,8 @@
 #define PROMPT_DELETE 	15
 #define PROMPT_SUPPRESS 16
 #define PROMPT_UNLINK	17
-#define PROMPT_NULL	18
+#define PROMPT_CLEAR_HTREE 18
+#define PROMPT_NULL	19
 
 /*
  * These are the prompts which are used to ask the user if they want
@@ -63,7 +64,8 @@ static const char *prompt[] = {
 	N_("Delete file"),	/* 15 */
 	N_("Suppress messages"),/* 16 */
 	N_("Unlink"),		/* 17 */
-	"",			/* 18 */
+	N_("Clear HTree index"),/* 18 */
+	"",			/* 19 */
 };
 
 /*
@@ -89,6 +91,8 @@ static const char *preen_msg[] = {
 	N_("FILE DELETED"),	/* 15 */
 	N_("SUPPRESSED"),	/* 16 */
 	N_("UNLINKED"),		/* 17 */
+	N_("HTREE INDEX CLEARED"),/* 18 */
+	"",			/* 19 */
 };
 
 static const struct e2fsck_problem problem_table[] = {
@@ -466,7 +470,7 @@ static const struct e2fsck_problem problem_table[] = {
 
 	/* Error allocating blocks for relocating metadata */
 	{ PR_1_RELOC_BLOCK_ALLOCATE,
-	  N_("@A %N @b(s) for %s: %m\n"),
+	  N_("@A %N contiguous @b(s) in @b @g %g for %s: %m\n"),
 	  PROMPT_NONE, PR_PREEN_OK },
 		
 	/* Error allocating block buffer during relocation process */
@@ -481,7 +485,7 @@ static const struct e2fsck_problem problem_table[] = {
 		
 	/* Relocating metatdata group information to X */
 	{ PR_1_RELOC_TO,
-	  N_("Relocating @g %g's %s to %c...\n"),
+	  N_("Relocating @g %g's %s to %c...\n"), /* xgettext:no-c-format */
 	  PROMPT_NONE, PR_PREEN_OK },
 		
 	/* Block read error during relocation process */
@@ -643,7 +647,56 @@ static const struct e2fsck_problem problem_table[] = {
 	/* Bad extended attribute value */
 	{ PR_1_EA_BAD_VALUE,
 	  N_("@a @b %b is corrupt (invalid value).  "),
-	  PROMPT_CLEAR, 0},	  
+	  PROMPT_CLEAR, 0},
+
+	/* Inode too big (latch question) */
+	{ PR_1_INODE_TOOBIG,
+	  N_("@i %i is too big.  "), PROMPT_TRUNCATE, 0 },
+
+	/* Directory too big */
+	{ PR_1_TOOBIG_DIR, 
+	  N_("@b #%B (%b) causes @d to be too big.  "),
+	  PROMPT_CLEAR, PR_LATCH_TOOBIG },
+
+	/* Regular file too big */
+	{ PR_1_TOOBIG_REG,
+	  N_("@b #%B (%b) causes file to be too big.  "),
+	  PROMPT_CLEAR, PR_LATCH_TOOBIG },
+
+	/* Symlink too big */
+	{ PR_1_TOOBIG_SYMLINK,
+	  N_("@b #%B (%b) causes symlink to be too big.  "),
+	  PROMPT_CLEAR, PR_LATCH_TOOBIG },
+
+	/* INDEX_FL flag set on a non-HTREE filesystem */
+	{ PR_1_HTREE_SET,
+	  N_("@i %i has INDEX_FL flag set on @f without htree support.\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* INDEX_FL flag set on a non-directory */	
+	{ PR_1_HTREE_NODIR,
+	  N_("@i %i has INDEX_FL flag set but is not a @d.\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* Invalid root node in HTREE directory */	
+	{ PR_1_HTREE_BADROOT,
+	  N_("@h %i has an invalid root node.\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* Unsupported hash version in HTREE directory */	
+	{ PR_1_HTREE_HASHV,
+	  N_("@h %i has an unsupported hash version (%N)\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* Incompatible flag in HTREE root node */	
+	{ PR_1_HTREE_INCOMPAT,
+	  N_("@h %i uses an incompatible htree root node flag.\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* HTREE too deep */	
+	{ PR_1_HTREE_DEPTH,
+	  N_("@h %i has a tree depth (%N) which is too big\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
 		  
 	/* Pass 1b errors */
 
@@ -682,7 +735,13 @@ static const struct e2fsck_problem problem_table[] = {
 	{ PR_1B_BLOCK_ITERATE,
 	  N_("Error while iterating over @bs in @i %i (%s): %m\n"),
 	  PROMPT_NONE, 0 },
-		  
+
+	/* Error adjusting EA refcount */
+	{ PR_1B_ADJ_EA_REFCOUNT,
+	  N_("Error addjusting refcount for @a @b %b (@i %i): %m\n"),
+	  PROMPT_NONE, 0 },
+
+
 	/* Pass 1C: Scan directories for inodes with dup blocks. */
 	{ PR_1C_PASS_HEADER,
 	  N_("Pass 1C: Scan directories for @is with dup @bs.\n"),
@@ -891,7 +950,7 @@ static const struct e2fsck_problem problem_table[] = {
 
 	/* Error iterating over directory blocks */
 	{ PR_2_DBLIST_ITERATE,
-	  N_("Error interating over @d @bs: %m\n"),
+	  N_("Error iterating over @d @bs: %m\n"),
 	  PROMPT_NONE, PR_FATAL },
 
 	/* Error reading directory block */
@@ -949,14 +1008,14 @@ static const struct e2fsck_problem problem_table[] = {
 	  N_("@E has a zero-length name.\n"),
 	  PROMPT_CLEAR, 0 },
 
-	/* Invalid fast symlink size */
-	{ PR_2_SYMLINK_SIZE,
-	  N_("Symlink %Q (@i #%i) has an invalid size (%Is).\n"),
+	/* Invalid symlink */
+	{ PR_2_INVALID_SYMLINK,
+	  N_("Symlink %Q (@i #%i) is invalid.\n"),
 	  PROMPT_CLEAR, 0 },
 
   	/* i_file_acl (extended attribute block) is bad */
 	{ PR_2_FILE_ACL_BAD,
-	  N_("@a @b @F is invalid (%If).\n"),
+	  N_("@a @b @F invalid (%If).\n"),
 	  PROMPT_CLEAR, 0 },
 
 	/* Filesystem contains large files, but has no such flag in sb */
@@ -964,6 +1023,85 @@ static const struct e2fsck_problem problem_table[] = {
 	  N_("@f contains large files, but lacks LARGE_FILE flag in @S.\n"),
 	  PROMPT_FIX, 0 },
 	  
+	/* Node in HTREE directory not referenced */
+	{ PR_2_HTREE_NOTREF,
+	  N_("@p @h %d: node (%B) not referenced\n"),
+	  PROMPT_NONE, 0 },
+
+	/* Node in HTREE directory referenced twice */
+	{ PR_2_HTREE_DUPREF,
+	  N_("@p @h %d: node (%B) referenced twice\n"),
+	  PROMPT_NONE, 0 },
+
+	/* Node in HTREE directory has bad min hash */
+	{ PR_2_HTREE_MIN_HASH,
+	  N_("@p @h %d: node (%B) has bad min hash\n"),
+	  PROMPT_NONE, 0 },
+
+	/* Node in HTREE directory has bad max hash */
+	{ PR_2_HTREE_MAX_HASH,
+	  N_("@p @h %d: node (%B) has bad max hash\n"),
+	  PROMPT_NONE, 0 },
+
+	/* Clear invalid HTREE directory */
+	{ PR_2_HTREE_CLEAR,
+	  N_("Invalid @h %d (%q).  "), PROMPT_CLEAR, 0 },
+		  
+	/* Clear the htree flag forcibly */
+	{ PR_2_HTREE_FCLR,
+	  N_("Forcibly clearing HTREE flag on @i %d (%q).  (Beta test code)\n"),
+		  PROMPT_NONE, 0 },
+
+	/* Bad block in htree interior node */
+	{ PR_2_HTREE_BADBLK,
+	  N_("@p @h %d (%q): bad @b number %b.\n"),
+	  PROMPT_CLEAR_HTREE, 0 },
+
+	/* Error adjusting EA refcount */
+	{ PR_2_ADJ_EA_REFCOUNT,
+	  N_("Error addjusting refcount for @a @b %b (@i %i): %m\n"),
+	  PROMPT_NONE, PR_FATAL },
+
+	/* Invalid HTREE root node */
+	{ PR_2_HTREE_BAD_ROOT,
+	  N_("@p @h %d: root node is invalid\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* Invalid HTREE limit */
+	{ PR_2_HTREE_BAD_LIMIT,
+	  N_("@p @h %d: node (%B) has bad limit (%N)\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* Invalid HTREE count */
+	{ PR_2_HTREE_BAD_COUNT,
+	  N_("@p @h %d: node (%B) has bad count (%N)\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* HTREE interior node has out-of-order hashes in table */
+	{ PR_2_HTREE_HASH_ORDER,
+	  N_("@p @h %d: node (%B) has an unordered hash table\n"),
+	  PROMPT_CLEAR_HTREE, PR_PREEN_OK },
+
+	/* Node in HTREE directory has bad depth */
+	{ PR_2_HTREE_BAD_DEPTH,
+	  N_("@p @h %d: node (%B) has bad depth\n"),
+	  PROMPT_NONE, 0 },
+	
+	/* Duplicate directory entry found */
+	{ PR_2_DUPLICATE_DIRENT,
+	  N_("Duplicate @E found.  "),
+	  PROMPT_CLEAR, 0 },
+	
+	/* Non-unique filename found */
+	{ PR_2_NON_UNIQUE_FILE, /* xgettext: no-c-format */
+	  N_("@E has a non-unique filename.\nRename to %s"),
+	  PROMPT_NULL, 0 },
+	
+	/* Duplicate directory entry found */
+	{ PR_2_REPORT_DUP_DIRENT,
+	  N_("Duplicate @e '%Dn' found.\n\tMarking %p (%i) to be rebuilt.\n\n"),
+	  PROMPT_NONE, 0 },
+	
 	/* Pass 3 errors */
 
 	/* Pass 3: Checking directory connectivity */
@@ -1084,7 +1222,39 @@ static const struct e2fsck_problem problem_table[] = {
 	/* Lost+found not a directory */
 	{ PR_3_LPF_NOTDIR,
 	  N_("/@l is not a @d (ino=%i)\n"),
-	  PROMPT_UNLINK, 0 }, 
+	  PROMPT_UNLINK, 0 },
+
+	/* Pass 3A Directory Optimization	*/
+
+	/* Pass 3A: Optimizing directories */
+	{ PR_3A_PASS_HEADER,
+	  N_("Pass 3A: Optimizing directories\n"),
+	  PROMPT_NONE, PR_PREEN_NOMSG },
+
+	/* Error iterating over directories */
+	{ PR_3A_OPTIMIZE_ITER,
+	  N_("Failed to create dirs_to_hash iterator: %m"),
+	  PROMPT_NONE, 0 },
+
+	/* Error rehash directory */
+	{ PR_3A_OPTIMIZE_DIR_ERR,
+	  N_("Failed to optimize directory %q (%d): %m"),
+	  PROMPT_NONE, 0 },
+
+	/* Rehashing dir header */
+	{ PR_3A_OPTIMIZE_DIR_HEADER,
+	  N_("Optimizing directories: "),
+	  PROMPT_NONE, PR_MSG_ONLY },
+
+	/* Rehashing directory %d */
+	{ PR_3A_OPTIMIZE_DIR,
+	  " %d",
+	  PROMPT_NONE, PR_LATCH_OPTIMIZE_DIR | PR_PREEN_NOHDR},
+		  
+	/* Rehashing dir end */	  
+	{ PR_3A_OPTIMIZE_DIR_END,
+	  "\n",
+	  PROMPT_NONE, PR_PREEN_NOHDR },
 
 	/* Pass 4 errors */
 	
@@ -1255,6 +1425,8 @@ static struct latch_descr pr_latch_info[] = {
 	{ PR_LATCH_RELOC, PR_0_RELOCATE_HINT, 0 },
 	{ PR_LATCH_DBLOCK, PR_1B_DUP_BLOCK_HEADER, PR_1B_DUP_BLOCK_END },
 	{ PR_LATCH_LOW_DTIME, PR_1_ORPHAN_LIST_REFUGEES, 0 },
+	{ PR_LATCH_TOOBIG, PR_1_INODE_TOOBIG, 0 },
+	{ PR_LATCH_OPTIMIZE_DIR, PR_3A_OPTIMIZE_DIR_HEADER, PR_3A_OPTIMIZE_DIR_END },
 	{ -1, 0, 0 },
 };
 
@@ -1371,13 +1543,10 @@ int fix_problem(e2fsck_t ctx, problem_t code, struct problem_context *pctx)
 		suppress++;
 	if (!suppress) {
 		message = ptr->e2p_description;
-		if (ctx->options & E2F_OPT_PREEN) {
+		if ((ctx->options & E2F_OPT_PREEN) &&
+		    !(ptr->flags & PR_PREEN_NOHDR)) {
 			printf("%s: ", ctx->device_name ?
 			       ctx->device_name : ctx->filesystem_name);
-#if 0
-			if (ptr->e2p_preen_msg)
-				message = ptr->e2p_preen_msg;
-#endif
 		}
 		print_e2fsck_message(ctx, _(message), pctx, 1);
 	}
