@@ -107,7 +107,7 @@ static void ext2fs_free_inode_cache(struct ext2_inode_cache *icache)
 /*
  * This procedure frees a badblocks list.
  */
-void ext2fs_badblocks_list_free(ext2_badblocks_list bb)
+void ext2fs_u32_list_free(ext2_u32_list bb)
 {
 	if (bb->magic != EXT2_ET_MAGIC_BADBLOCKS_LIST)
 		return;
@@ -118,6 +118,12 @@ void ext2fs_badblocks_list_free(ext2_badblocks_list bb)
 	ext2fs_free_mem((void **) &bb);
 }
 
+void ext2fs_badblocks_list_free(ext2_badblocks_list bb)
+{
+	ext2fs_u32_list_free((ext2_u32_list) bb);
+}
+
+
 /*
  * Free a directory block list
  */
