@@ -2,7 +2,10 @@
  * getsize.c --- get the size of a partition.
  * 
  * Copyright (C) 1995, 1995 Theodore Ts'o.
+ * Copyright (C) 2003 VMware, Inc.
  *
+ * Windows version of ext2fs_get_device_size by Chris Li, VMware.
+ * 
  * %Begin-Header%
  * This file may be redistributed under the terms of the GNU Public
  * License.
@@ -33,9 +36,62 @@
 #define BLKGETSIZE _IO(0x12,96)	/* return device size */
 #endif
 
+#ifdef APPLE_DARWIN
+#include <sys/ioctl.h>
+#include <sys/disk.h>
+
+#define BLKGETSIZE DKIOCGETBLOCKCOUNT32
+#endif /* APPLE_DARWIN */
+
 #include "ext2_fs.h"
 #include "ext2fs.h"
 
+#if defined(__CYGWIN__) || defined (WIN32)
+#include "windows.h"
+#include "winioctl.h"
+
+errcode_t ext2fs_get_device_size(const char *file, int blocksize,
+				 blk_t *retblocks)
+{
+	HANDLE dev;
+	PARTITION_INFORMATION pi;
+	DISK_GEOMETRY gi;
+	DWORD retbytes;
+	LARGE_INTEGER filesize;
+
+	dev = CreateFile(file, GENERIC_READ, 
+			 FILE_SHARE_READ | FILE_SHARE_WRITE ,
+                	 NULL,  OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,  NULL); 
+ 
+	if (dev == INVALID_HANDLE_VALUE)
+		return EBADF;
+	if (DeviceIoControl(dev, IOCTL_DISK_GET_PARTITION_INFO,
+			    &pi, sizeof(PARTITION_INFORMATION),
+			    &pi, sizeof(PARTITION_INFORMATION),
+			    &retbytes, NULL)) {
+
+		*retblocks = pi.PartitionLength.QuadPart / blocksize;
+	
+	} else if (DeviceIoControl(dev, IOCTL_DISK_GET_DRIVE_GEOMETRY,
+				&gi, sizeof(DISK_GEOMETRY),
+				&gi, sizeof(DISK_GEOMETRY),
+				&retbytes, NULL)) {
+
+		*retblocks = gi.BytesPerSector *
+			     gi.SectorsPerTrack *
+			     gi.TracksPerCylinder *
+			     gi.Cylinders.QuadPart / blocksize;
+
+	} else if (GetFileSizeEx(dev, &filesize)) {
+		*retblocks = filesize.QuadPart / blocksize;
+	}
+
+	CloseHandle(dev);
+	return 0;
+}
+
+#else
+
 static int valid_offset (int fd, ext2_loff_t offset)
 {
 	char ch;
@@ -134,6 +190,8 @@ errcode_t ext2fs_get_device_size(const char *file, int blocksize,
 	return 0;
 }
 
+#endif /* WIN32 */
+
 #ifdef DEBUG
 int main(int argc, char **argv)
 {
