@@ -9,6 +9,9 @@
  * %End-Header%
  */
 
+#define _LARGEFILE_SOURCE
+#define _LARGEFILE64_SOURCE
+
 #if HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
@@ -27,10 +30,12 @@
 
 #ifdef __linux__
 
-#ifdef HAVE_LLSEEK
-#if HAVE_UNISTD_H
-#include <unistd.h>
-#endif
+#if defined(HAVE_LSEEK64) && defined(HAVE_LSEEK64_PROTOTYPE)
+
+#define my_llseek lseek64
+
+#else
+#if defined(HAVE_LLSEEK)
 #include <syscall.h>
 
 #ifndef HAVE_LLSEEK_PROTOTYPE
@@ -39,13 +44,13 @@ extern long long llseek (int fd, long long offset, int origin);
 
 #define my_llseek llseek
 
-#else	/* HAVE_LLSEEK */
+#else	/* ! HAVE_LLSEEK */
 
-#ifdef __alpha__
+#if defined(__alpha__) || defined (__ia64__)
 
 #define llseek lseek
 
-#else /* !__alpha__ */
+#else /* !__alpha__ && !__ia64__*/
 
 #include <linux/unistd.h>
 
@@ -77,9 +82,10 @@ static ext2_loff_t my_llseek (int fd, ext2_loff_t offset, int origin)
 	return (retval == -1 ? (ext2_loff_t) retval : result);
 }
 
-#endif	/* HAVE_LLSEEK */
+#endif	/* __alpha__ || __ia64__ */
 
-#endif /* __alpha__ */
+#endif /* HAVE_LLSEEK */
+#endif /* defined(HAVE_LSEEK64) && defined(HAVE_LSEEK64_PROTOTYPE) */
 
 ext2_loff_t ext2fs_llseek (int fd, ext2_loff_t offset, int origin)
 {
