@@ -23,15 +23,11 @@
 #include <sys/types.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
+#include "e2image.h"
 
-#ifdef __powerpc__
+#if defined(__powerpc__) && defined(EXT2FS_ENABLE_SWAPFS)
 /*
  * On the PowerPC, the big-endian variant of the ext2 filesystem
  * has its bitmaps stored as 32-bit words with bit 0 as the LSB
@@ -45,7 +41,7 @@
 #endif
 
 #ifdef EXT2_BIG_ENDIAN_BITMAPS
-void ext2fs_swap_bitmap(ext2_filsys fs, char *bitmap, int nbytes)
+static void ext2fs_swap_bitmap(ext2_filsys fs, char *bitmap, int nbytes)
 {
 	__u32 *p = (__u32 *) bitmap;
 	int n;
@@ -92,7 +88,6 @@ errcode_t ext2fs_write_inode_bitmap(ext2_filsys fs)
 		}
 		inode_bitmap += nbytes;
 	}
-	fs->flags |= EXT2_FLAG_CHANGED;
 	fs->flags &= ~EXT2_FLAG_IB_DIRTY;
 	ext2fs_free_mem((void **) &bitmap_block);
 	return 0;
@@ -145,7 +140,6 @@ errcode_t ext2fs_write_block_bitmap (ext2_filsys fs)
 		}
 		block_bitmap += nbytes;
 	}
-	fs->flags |= EXT2_FLAG_CHANGED;
 	fs->flags &= ~EXT2_FLAG_BB_DIRTY;
 	ext2fs_free_mem((void **) &bitmap_block);
 	return 0;
@@ -188,6 +182,28 @@ static errcode_t read_bitmaps(ext2_filsys fs, int do_inode, int do_block)
 	}
 	ext2fs_free_mem((void **) &buf);
 
+	if (fs->flags & EXT2_FLAG_IMAGE_FILE) {
+		if (inode_bitmap) {
+			blk = (fs->image_header->offset_inodemap /
+			       fs->blocksize);
+			retval = io_channel_read_blk(fs->io, blk,
+			     -(inode_nbytes * fs->group_desc_count),
+			     inode_bitmap);
+			if (retval)
+				goto cleanup;
+		}
+		if (block_bitmap) {
+			blk = (fs->image_header->offset_blockmap /
+			       fs->blocksize);
+			retval = io_channel_read_blk(fs->io, blk, 
+			     -(block_nbytes * fs->group_desc_count),
+			     block_bitmap);
+			if (retval)
+				goto cleanup;
+		}
+		return 0;
+	}
+
 	for (i = 0; i < fs->group_desc_count; i++) {
 		if (block_bitmap) {
 			blk = fs->group_desc[i].bg_block_bitmap;
