@@ -36,6 +36,9 @@
 struct test_private_data {
 	int	magic;
 	io_channel real;
+	int flags;
+	FILE *outfile;
+	unsigned long block;
 	void (*read_blk)(unsigned long block, int count, errcode_t err);
 	void (*write_blk)(unsigned long block, int count, errcode_t err);
 	void (*set_blksize)(int blksize, errcode_t err);
@@ -82,6 +85,34 @@ void (*test_io_cb_set_blksize)
 void (*test_io_cb_write_byte)
 	(unsigned long block, int count, errcode_t err) = 0;
 
+/*
+ * Test flags
+ */
+#define TEST_FLAG_READ			0x01
+#define TEST_FLAG_WRITE			0x02
+#define TEST_FLAG_SET_BLKSIZE		0x04
+#define TEST_FLAG_FLUSH			0x08
+
+static void test_dump_block(io_channel channel,
+			    struct test_private_data *data,
+			    unsigned long block, const void *buf)
+{
+	const unsigned char *cp;
+	FILE *f = data->outfile;
+	int	i;
+	unsigned long	cksum = 0;
+
+	for (i=0, cp = buf; i < channel->block_size; i++, cp++) {
+		cksum += *cp;
+	}
+	fprintf(f, "Contents of block %lu, checksum %08lu: \n", block, cksum);
+	for (i=0, cp = buf; i < channel->block_size; i++, cp++) {
+		if ((i % 16) == 0)
+			fprintf(f, "%04x: ", i);
+		fprintf(f, "%02x%c", *cp, ((i % 16) == 15) ? '\n' : ' ');
+	}
+}
+
 static errcode_t test_open(const char *name, int flags, io_channel *channel)
 {
 	io_channel	io = NULL;
@@ -127,6 +158,20 @@ static errcode_t test_open(const char *name, int flags, io_channel *channel)
 	data->write_blk = 	test_io_cb_write_blk;
 	data->set_blksize = 	test_io_cb_set_blksize;
 	data->write_byte = 	test_io_cb_write_byte;
+
+	data->outfile = NULL;
+	if (getenv("TEST_IO_LOGFILE"))
+		data->outfile = fopen(getenv("TEST_IO_LOGFILE"), "w");
+	if (!data->outfile)
+		data->outfile = stderr;
+
+	data->flags = 0;
+	if (getenv("TEST_IO_FLAGS"))
+		data->flags = strtoul(getenv("TEST_IO_FLAGS"), NULL, 0);
+	
+	data->block = 0;
+	if (getenv("TEST_IO_BLOCK"))
+		data->block = strtoul(getenv("TEST_IO_BLOCK"), NULL, 0);
 	
 	*channel = io;
 	return 0;
@@ -153,6 +198,9 @@ static errcode_t test_close(io_channel channel)
 	
 	if (data->real)
 		retval = io_channel_close(data->real);
+
+	if (data->outfile && data->outfile != stderr)
+		fclose(data->outfile);
 	
 	ext2fs_free_mem((void **) &channel->private_data);
 	if (channel->name)
@@ -174,9 +222,11 @@ static errcode_t test_set_blksize(io_channel channel, int blksize)
 		retval = io_channel_set_blksize(data->real, blksize);
 	if (data->set_blksize)
 		data->set_blksize(blksize, retval);
-	else
-		printf("Test_io: set_blksize(%d) returned %s\n",
-		       blksize, retval ? error_message(retval) : "OK");
+	if (data->flags & TEST_FLAG_SET_BLKSIZE)
+		fprintf(data->outfile,
+			"Test_io: set_blksize(%d) returned %s\n",
+			blksize, retval ? error_message(retval) : "OK");
+	channel->block_size = blksize;
 	return retval;
 }
 
@@ -195,9 +245,12 @@ static errcode_t test_read_blk(io_channel channel, unsigned long block,
 		retval = io_channel_read_blk(data->real, block, count, buf);
 	if (data->read_blk)
 		data->read_blk(block, count, retval);
-	else
-		printf("Test_io: read_blk(%lu, %d) returned %s\n",
-		       block, count, retval ? error_message(retval) : "OK");
+	if (data->flags & TEST_FLAG_READ)
+		fprintf(data->outfile,
+			"Test_io: read_blk(%lu, %d) returned %s\n",
+			block, count, retval ? error_message(retval) : "OK");
+	if (data->block && data->block == block)
+		test_dump_block(channel, data, block, buf);
 	return retval;
 }
 
@@ -215,9 +268,12 @@ static errcode_t test_write_blk(io_channel channel, unsigned long block,
 		retval = io_channel_write_blk(data->real, block, count, buf);
 	if (data->write_blk)
 		data->write_blk(block, count, retval);
-	else
-		printf("Test_io: write_blk(%lu, %d) returned %s\n",
-		       block, count, retval ? error_message(retval) : "OK");
+	if (data->flags & TEST_FLAG_WRITE)
+		fprintf(data->outfile,
+			"Test_io: write_blk(%lu, %d) returned %s\n",
+			block, count, retval ? error_message(retval) : "OK");
+	if (data->block && data->block == block)
+		test_dump_block(channel, data, block, buf);
 	return retval;
 }
 
@@ -235,9 +291,10 @@ static errcode_t test_write_byte(io_channel channel, unsigned long offset,
 		retval = io_channel_write_byte(data->real, offset, count, buf);
 	if (data->write_byte)
 		data->write_byte(offset, count, retval);
-	else
-		printf("Test_io: write_byte(%lu, %d) returned %s\n",
-		       offset, count, retval ? error_message(retval) : "OK");
+	if (data->flags & TEST_FLAG_WRITE)
+		fprintf(data->outfile,
+			"Test_io: write_byte(%lu, %d) returned %s\n",
+			offset, count, retval ? error_message(retval) : "OK");
 	return retval;
 }
 
@@ -256,8 +313,9 @@ static errcode_t test_flush(io_channel channel)
 	if (data->real)
 		retval = io_channel_flush(data->real);
 	
-	printf("Test_io: flush() returned %s\n",
-		       retval ? error_message(retval) : "OK");
+	if (data->flags & TEST_FLAG_FLUSH)
+		fprintf(data->outfile, "Test_io: flush() returned %s\n",
+			retval ? error_message(retval) : "OK");
 	
 	return retval;
 }
