@@ -1,8 +1,11 @@
 /*
- * unix_io.c --- This is the Unix I/O interface to the I/O manager.
+ * unix_io.c --- This is the Unix (well, really POSIX) implementation
+ * 	of the I/O manager.
  *
  * Implements a one-block write-through cache.
  *
+ * Includes support for Windows NT support under Cygwin. 
+ *
  * Copyright (C) 1993, 1994, 1995, 1996, 1997, 1998, 1999, 2000, 2001,
  * 	2002 by Theodore Ts'o.
  *
@@ -34,7 +37,9 @@
 #if HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
+#if HAVE_SYS_RESOURCE_H
 #include <sys/resource.h>
+#endif
 
 #include "ext2_fs.h"
 #include "ext2fs.h"
@@ -77,6 +82,9 @@ static errcode_t unix_flush(io_channel channel);
 static errcode_t unix_write_byte(io_channel channel, unsigned long offset,
 				int size, const void *data);
 
+static void reuse_cache(io_channel channel, struct unix_private_data *data,
+		 struct unix_cache *cache, unsigned long block);
+
 static struct struct_io_manager struct_unix_manager = {
 	EXT2_ET_MAGIC_IO_MANAGER,
 	"Unix I/O Manager",
@@ -86,7 +94,11 @@ static struct struct_io_manager struct_unix_manager = {
 	unix_read_blk,
 	unix_write_blk,
 	unix_flush,
+#ifdef __CYGWIN__
+	0
+#else
 	unix_write_byte
+#endif
 };
 
 io_manager unix_io_manager = &struct_unix_manager;
@@ -94,6 +106,7 @@ io_manager unix_io_manager = &struct_unix_manager;
 /*
  * Here are the raw I/O functions
  */
+#ifndef __CYGWIN__
 static errcode_t raw_read_blk(io_channel channel,
 			      struct unix_private_data *data,
 			      unsigned long block,
@@ -126,6 +139,60 @@ error_out:
 					       size, actual, retval);
 	return retval;
 }
+#else /* __CYGWIN__ */
+/*
+ * Windows block devices only allow sector alignment IO in offset and size
+ */
+static errcode_t raw_read_blk(io_channel channel,
+			      struct unix_private_data *data,
+			      unsigned long block,
+			      int count, void *buf)
+{
+	errcode_t	retval;
+	size_t		size, alignsize, fragment;
+	ext2_loff_t	location;
+	int		total = 0, actual;
+#define BLOCKALIGN 512
+	char		sector[BLOCKALIGN];
+
+	size = (count < 0) ? -count : count * channel->block_size;
+	location = (ext2_loff_t) block * channel->block_size;
+#ifdef DEBUG
+	printf("count=%d, size=%d, block=%d, blk_size=%d, location=%lx\n",
+	 		count, size, block, channel->block_size, location);
+#endif
+	if (ext2fs_llseek(data->dev, location, SEEK_SET) != location) {
+		retval = errno ? errno : EXT2_ET_LLSEEK_FAILED;
+		goto error_out;
+	}
+	fragment = size % BLOCKALIGN;
+	alignsize = size - fragment;
+	if (alignsize) {
+		actual = read(data->dev, buf, alignsize);
+		if (actual != alignsize)
+			goto short_read;
+	}
+	if (fragment) {
+		actual = read(data->dev, sector, BLOCKALIGN);
+		if (actual != BLOCKALIGN)
+			goto short_read;
+		memcpy(buf+alignsize, sector, fragment);
+	}
+	return 0;
+
+short_read:
+	if (actual>0)
+		total += actual;
+	retval = EXT2_ET_SHORT_READ;
+
+error_out:
+	memset((char *) buf+total, 0, size-actual);
+	if (channel->read_error)
+		retval = (channel->read_error)(channel, block, count, buf,
+					       size, actual, retval);
+	return retval;
+}
+#endif
 
 static errcode_t raw_write_blk(io_channel channel,
 			       struct unix_private_data *data,
@@ -211,6 +278,7 @@ static void free_cache(io_channel channel,
 	}
 }
 
+#ifndef NO_IO_CACHE
 /*
  * Try to find a block in the cache.  If the block is not found, and
  * eldest is a non-zero pointer, then fill in eldest with the cache
@@ -247,7 +315,7 @@ static struct unix_cache *find_cached_block(io_channel channel,
 /*
  * Reuse a particular cache entry for another block.
  */
-void reuse_cache(io_channel channel, struct unix_private_data *data,
+static void reuse_cache(io_channel channel, struct unix_private_data *data,
 		 struct unix_cache *cache, unsigned long block)
 {
 	if (cache->dirty && cache->in_use)
@@ -291,8 +359,7 @@ static errcode_t flush_cached_blocks(io_channel channel,
 	}
 	return retval2;
 }
-
-
+#endif /* NO_IO_CACHE */
 
 static errcode_t unix_open(const char *name, int flags, io_channel *channel)
 {
@@ -406,7 +473,9 @@ static errcode_t unix_close(io_channel channel)
 	if (--channel->refcount > 0)
 		return 0;
 
+#ifndef NO_IO_CACHE
 	retval = flush_cached_blocks(channel, data, 0);
+#endif
 
 	if (close(data->dev) < 0)
 		retval = errno;
@@ -429,8 +498,10 @@ static errcode_t unix_set_blksize(io_channel channel, int blksize)
 	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);
 
 	if (channel->block_size != blksize) {
+#ifndef NO_IO_CACHE
 		if ((retval = flush_cached_blocks(channel, data, 0)))
 			return retval;
+#endif
 		
 		channel->block_size = blksize;
 		free_cache(channel, data);
@@ -454,6 +525,9 @@ static errcode_t unix_read_blk(io_channel channel, unsigned long block,
 	data = (struct unix_private_data *) channel->private_data;
 	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);
 
+#ifdef NO_IO_CACHE
+	return raw_read_blk(channel, data, block, count, buf);
+#else
 	/*
 	 * If we're doing an odd-sized read or a very large read,
 	 * flush out the cache and then do a direct read.
@@ -502,6 +576,7 @@ static errcode_t unix_read_blk(io_channel channel, unsigned long block,
 		}
 	}
 	return 0;
+#endif /* NO_IO_CACHE */
 }
 
 static errcode_t unix_write_blk(io_channel channel, unsigned long block,
@@ -509,7 +584,7 @@ static errcode_t unix_write_blk(io_channel channel, unsigned long block,
 {
 	struct unix_private_data *data;
 	struct unix_cache *cache, *reuse;
-	errcode_t	retval = 0, retval2;
+	errcode_t	retval = 0;
 	const char	*cp;
 	int		writethrough;
 
@@ -517,6 +592,9 @@ static errcode_t unix_write_blk(io_channel channel, unsigned long block,
 	data = (struct unix_private_data *) channel->private_data;
 	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);
 
+#ifdef NO_IO_CACHE
+	return raw_write_blk(channel, data, block, count, buf);
+#else	
 	/*
 	 * If we're doing an odd-sized write or a very large write,
 	 * flush out the cache completely and then do a direct write.
@@ -550,6 +628,7 @@ static errcode_t unix_write_blk(io_channel channel, unsigned long block,
 		cp += channel->block_size;
 	}
 	return retval;
+#endif /* NO_IO_CACHE */
 }
 
 static errcode_t unix_write_byte(io_channel channel, unsigned long offset,
@@ -563,11 +642,13 @@ static errcode_t unix_write_byte(io_channel channel, unsigned long offset,
 	data = (struct unix_private_data *) channel->private_data;
 	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);
 
+#ifndef NO_IO_CACHE
 	/*
 	 * Flush out the cache completely
 	 */
 	if ((retval = flush_cached_blocks(channel, data, 1)))
 		return retval;
+#endif
 
 	if (lseek(data->dev, offset, SEEK_SET) < 0)
 		return errno;
@@ -591,7 +672,9 @@ static errcode_t unix_flush(io_channel channel)
 	data = (struct unix_private_data *) channel->private_data;
 	EXT2_CHECK_MAGIC(data, EXT2_ET_MAGIC_UNIX_IO_CHANNEL);
 
+#ifndef NO_IO_CACHE
 	retval = flush_cached_blocks(channel, data, 0);
+#endif
 	fsync(data->dev);
 	return retval;
 }
