@@ -16,12 +16,7 @@
 #include <stdio.h>
 #include <ctype.h>
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
 
 #include "../../version.h"
