@@ -41,12 +41,8 @@ void ss_list_requests(argc, argv, sci_idx, info_ptr)
     char buffer[BUFSIZ];
     FILE *output;
     int fd;
-#ifdef POSIX_SIGNALS
     sigset_t omask, igmask;
-#else
-    int mask;
-#endif
-    sigret_t (*func) PROTOTYPE((int));
+    sigret_t (*func)(int);
 #ifndef NO_FORK
     int waitb;
 #endif
@@ -54,21 +50,13 @@ void ss_list_requests(argc, argv, sci_idx, info_ptr)
     DONT_USE(argc);
     DONT_USE(argv);
 
-#ifdef POSIX_SIGNALS
     sigemptyset(&igmask);
     sigaddset(&igmask, SIGINT);
     sigprocmask(SIG_BLOCK, &igmask, &omask);
-#else
-    mask = sigblock(sigmask(SIGINT));
-#endif
     func = signal(SIGINT, SIG_IGN);
     fd = ss_pager_create();
     output = fdopen(fd, "w");
-#ifdef POSIX_SIGNALS
     sigprocmask(SIG_SETMASK, &omask, (sigset_t *) 0);
-#else
-    sigsetmask(mask);
-#endif
 
     fprintf (output, "Available %s requests:\n\n",
 	     ss_info (sci_idx) -> subsystem_name);
