@@ -1,7 +1,7 @@
 /*
  * prompt.c: Routines for retrieving and setting a prompt.
  *
- * $Header: lib/ss/SCCS/s.prompt.c 1.15 99/10/23 01:16:22-00:00 tytso@mit.edu $
+ * $Header: lib/ss/SCCS/s.prompt.c 1.16 03/03/30 22:26:12-05:00 tytso@think.thunk.org $
  * $Locker: <Not implemented> $
  *
  * Copyright 1987, 1988 by MIT Student Information Processing Board
@@ -19,23 +19,12 @@
 #include <stdio.h>
 #include "ss_internal.h"
 
-#ifdef __STDC__
 void ss_set_prompt(int sci_idx, char *new_prompt)
-#else
-void ss_set_prompt(sci_idx, new_prompt)
-     int sci_idx;
-     char *new_prompt;
-#endif
 {
      ss_info(sci_idx)->prompt = new_prompt;
 }
 
-#ifdef __STDC__
 char *ss_get_prompt(int sci_idx)
-#else
-char *ss_get_prompt(sci_idx)
-     int sci_idx;
-#endif
 {
      return(ss_info(sci_idx)->prompt);
 }
