@@ -86,24 +86,25 @@ static void list_desc (ext2_filsys fs)
 	unsigned long i;
 	long diff;
 	blk_t	group_blk, next_blk;
-	char * block_bitmap = fs->block_map->bitmap;
-	char * inode_bitmap = fs->inode_map->bitmap;
-	int inode_blocks_per_group;
-	int group_desc_blocks;
+	char *block_bitmap=NULL, *inode_bitmap=NULL;
+	int inode_blocks_per_group, old_desc_blocks;
+	int meta_bg, meta_bg_size, has_super;
+
+	if (fs->block_map)
+		block_bitmap = fs->block_map->bitmap;
+	if (fs->inode_map)
+		inode_bitmap = fs->inode_map->bitmap;
 
 	inode_blocks_per_group = ((fs->super->s_inodes_per_group *
 				   EXT2_INODE_SIZE(fs->super)) +
 				  EXT2_BLOCK_SIZE(fs->super) - 1) /
 				 EXT2_BLOCK_SIZE(fs->super);
-	group_desc_blocks = ((fs->super->s_blocks_count -
-			      fs->super->s_first_data_block +
-			      EXT2_BLOCKS_PER_GROUP(fs->super) - 1) /
-			     EXT2_BLOCKS_PER_GROUP(fs->super) +
-			     EXT2_DESC_PER_BLOCK(fs->super) - 1) /
-			    EXT2_DESC_PER_BLOCK(fs->super);
-
 	fputc('\n', stdout);
 	group_blk = fs->super->s_first_data_block;
+	if (fs->super->s_feature_incompat & EXT2_FEATURE_INCOMPAT_META_BG)
+		old_desc_blocks = fs->super->s_first_meta_bg;
+	else
+		old_desc_blocks = fs->desc_blocks;
 	for (i = 0; i < fs->group_desc_count; i++) {
 		next_blk = group_blk + fs->super->s_blocks_per_group;
 		if (next_blk > fs->super->s_blocks_count)
@@ -111,14 +112,36 @@ static void list_desc (ext2_filsys fs)
 		printf (_("Group %lu: (Blocks "), i);
 		printf(range_format, group_blk, next_blk - 1);
 		fputs(")\n", stdout);
-		if (ext2fs_bg_has_super (fs, i)) {
-			printf (_("  %s Superblock at "),
+		has_super = ext2fs_bg_has_super(fs, i);
+		if (has_super) {
+			printf (_("  %s superblock at "),
 				i == 0 ? _("Primary") : _("Backup"));
 			printf(num_format, group_blk);
-			printf(_(",  Group Descriptors at "));
-			printf(range_format, group_blk+1,
-			       group_blk + group_desc_blocks);
-			fputc('\n', stdout);
+		}
+		meta_bg_size = (fs->blocksize /
+				sizeof (struct ext2_group_desc));
+		meta_bg = i / meta_bg_size;
+		if (!(fs->super->s_feature_incompat &
+		      EXT2_FEATURE_INCOMPAT_META_BG) ||
+		    (meta_bg < fs->super->s_first_meta_bg)) {
+			if (has_super) {
+				printf(_(", Group descriptors at "));
+				printf(range_format, group_blk+1,
+				       group_blk + old_desc_blocks);
+				fputc('\n', stdout);
+			}
+		} else {
+			if (has_super)
+				has_super = 1;
+			if (((i % meta_bg_size) == 0) ||
+			    ((i % meta_bg_size) == 1) ||
+			    ((i % meta_bg_size) == (meta_bg_size-1))) {
+				fputc(has_super ? ',' : ' ', stdout);
+				printf(_(" Group descriptor at "));
+				printf(num_format, group_blk + has_super);
+				fputc('\n', stdout);
+			} else if (has_super)
+				fputc('\n', stdout);
 		}
 		fputs(_("  Block bitmap at "), stdout);
 		printf(num_format, fs->group_desc[i].bg_block_bitmap);
@@ -138,17 +161,25 @@ static void list_desc (ext2_filsys fs)
 		if (diff > 0)
 			printf(" (+%ld)", diff);
 		printf (_("\n  %d free blocks, %d free inodes, "
-			  "%d directories\n  Free blocks: "),
+			  "%d directories\n"),
 			fs->group_desc[i].bg_free_blocks_count,
 			fs->group_desc[i].bg_free_inodes_count,
 			fs->group_desc[i].bg_used_dirs_count);
-		print_free (i, block_bitmap, fs->super->s_blocks_per_group,
-			    fs->super->s_first_data_block);
-		fputs(_("\n  Free inodes: "), stdout);
-		print_free (i, inode_bitmap, fs->super->s_inodes_per_group, 1);
-		fputc('\n', stdout);
-		block_bitmap += fs->super->s_blocks_per_group / 8;
-		inode_bitmap += fs->super->s_inodes_per_group / 8;
+		if (block_bitmap) {
+			fputs(_("  Free blocks: "), stdout);
+			print_free (i, block_bitmap,
+				    fs->super->s_blocks_per_group,
+				    fs->super->s_first_data_block);
+			fputc('\n', stdout);
+			block_bitmap += fs->super->s_blocks_per_group / 8;
+		}
+		if (inode_bitmap) {
+			fputs(_("  Free inodes: "), stdout);
+			print_free (i, inode_bitmap,
+				    fs->super->s_inodes_per_group, 1);
+			fputc('\n', stdout);
+			inode_bitmap += fs->super->s_inodes_per_group / 8;
+		}
 		group_blk = next_blk;
 	}
 }
@@ -330,14 +361,12 @@ int main (int argc, char ** argv)
 			exit (0);
 		}
 		retval = ext2fs_read_bitmaps (fs);
-		if (retval) {
-			com_err (program_name, retval,
-				 _("while trying to read the bitmaps"),
-				 device_name);
-			ext2fs_close (fs);
-			exit (1);
-		}
 		list_desc (fs);
+		if (retval) {
+			printf(_("\n%s: %s: error reading bitmaps: %s\n"),
+			       program_name, device_name,
+			       error_message(retval));
+		}
 	}
 	ext2fs_close (fs);
 	exit (0);
