@@ -147,11 +147,13 @@ struct mke2fs_defaults {
 
 static void set_fs_defaults(const char *fs_type,
 			    struct ext2_super_block *super,
-			    int blocksize, int *inode_ratio)
+			    int blocksize, int sector_size,
+			    int *inode_ratio)
 {
 	int	megs;
 	int	ratio = 0;
 	struct mke2fs_defaults *p;
+	int	use_bsize = 1024;
 
 	megs = super->s_blocks_count * (EXT2_BLOCK_SIZE(super) / 1024) / 1024;
 	if (inode_ratio)
@@ -162,23 +164,28 @@ static void set_fs_defaults(const char *fs_type,
 		if ((strcmp(p->type, fs_type) != 0) &&
 		    (strcmp(p->type, default_str) != 0))
 			continue;
-		if ((p->size != 0) &&
-		    (megs > p->size))
+		if ((p->size != 0) && (megs > p->size))
 			continue;
 		if (ratio == 0)
 			*inode_ratio = p->inode_ratio < blocksize ?
 				blocksize : p->inode_ratio;
-		if (blocksize == 0) {
-			if (p->blocksize == DEF_MAX_BLOCKSIZE)
-				p->blocksize = sys_page_size;
-			super->s_log_frag_size = super->s_log_block_size =
-				int_log2(p->blocksize >> EXT2_MIN_BLOCK_LOG_SIZE);
-		}
+		use_bsize = p->blocksize;
 	}
-	if (blocksize == 0)
-		super->s_blocks_count /= EXT2_BLOCK_SIZE(super) / 1024;
+	if (sector_size && use_bsize < sector_size)
+		use_bsize = sector_size;
+	if (blocksize <= 0) {
+		if (use_bsize == DEF_MAX_BLOCKSIZE)
+			use_bsize = sys_page_size;
+		if ((blocksize < 0) && (use_bsize < (-blocksize)))
+			use_bsize = -blocksize;
+		blocksize = use_bsize;
+		super->s_blocks_count /= blocksize / 1024;
+	}
+	super->s_log_frag_size = super->s_log_block_size =
+		int_log2(blocksize >> EXT2_MIN_BLOCK_LOG_SIZE);
 }
 
+
 /*
  * Helper function for read_bb_file and test_disk
  */
@@ -788,14 +795,14 @@ static __u32 ok_features[3] = {
 
 static void PRS(int argc, char *argv[])
 {
-	int		c;
+	int		b, c;
 	int		size;
 	char *		tmp;
-	blk_t		group_blk_max = 8192;
 	int		blocksize = 0;
 	int		inode_ratio = 0;
 	int		inode_size = 0;
 	int		reserved_ratio = 5;
+	int		sector_size = 0;
 	ext2_ino_t	num_inodes = 0;
 	errcode_t	retval;
 	char *		oldpath = getenv("PATH");
@@ -843,7 +850,9 @@ static void PRS(int argc, char *argv[])
 	param.s_rev_level = 1;  /* Create revision 1 filesystems now */
 	param.s_feature_incompat |= EXT2_FEATURE_INCOMPAT_FILETYPE;
 	param.s_feature_ro_compat |= EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
+#if 0
 	param.s_feature_compat |= EXT2_FEATURE_COMPAT_DIR_INDEX;
+#endif
 
 #ifdef __linux__
 	if (uname(&ut)) {
@@ -859,8 +868,6 @@ static void PRS(int argc, char *argv[])
 		param.s_feature_ro_compat = 0;
 	}
 #endif
-	fprintf (stderr, "mke2fs %s (%s)\n",
-		 E2FSPROGS_VERSION, E2FSPROGS_DATE);
 
 	if (argc && *argv) {
 		program_name = get_progname(*argv);
@@ -874,9 +881,10 @@ static void PRS(int argc, char *argv[])
 		    "b:cf:g:i:jl:m:no:qr:R:s:tvI:J:ST:FL:M:N:O:V")) != EOF)
 		switch (c) {
 		case 'b':
-			blocksize = strtoul(optarg, &tmp, 0);
-			if (blocksize < EXT2_MIN_BLOCK_SIZE ||
-			    blocksize > EXT2_MAX_BLOCK_SIZE || *tmp) {
+			blocksize = strtol(optarg, &tmp, 0);
+			b = (blocksize > 0) ? blocksize : -blocksize;
+			if (b < EXT2_MIN_BLOCK_SIZE ||
+			    b > EXT2_MAX_BLOCK_SIZE || *tmp) {
 				com_err(program_name, 0,
 					_("bad block size - %s"), optarg);
 				exit(1);
@@ -885,9 +893,10 @@ static void PRS(int argc, char *argv[])
 				fprintf(stderr, _("Warning: blocksize %d not "
 						  "usable on most systems.\n"),
 					blocksize);
-			param.s_log_block_size =
-				int_log2(blocksize >> EXT2_MIN_BLOCK_LOG_SIZE);
-			group_blk_max = blocksize * 8;
+			if (blocksize > 0) 
+				param.s_log_block_size =
+					int_log2(blocksize >>
+						 EXT2_MIN_BLOCK_LOG_SIZE);
 			break;
 		case 'c':	/* Check for bad blocks */
 		case 't':	/* deprecated */
@@ -1060,6 +1069,10 @@ static void PRS(int argc, char *argv[])
 	if (optind < argc)
 		usage();
 
+	if (!quiet)
+		fprintf (stderr, "mke2fs %s (%s)\n", E2FSPROGS_VERSION, 
+			 E2FSPROGS_DATE);
+
 	if (raid_opts)
 		parse_raid_opts(raid_opts);
 
@@ -1067,18 +1080,32 @@ static void PRS(int argc, char *argv[])
 	 * If there's no blocksize specified and there is a journal
 	 * device, use it to figure out the blocksize
 	 */
-	if (blocksize == 0 && journal_device) {
+	if (blocksize <= 0 && journal_device) {
 		ext2_filsys	jfs;
+		io_manager	io_ptr;
 
+#ifdef CONFIG_TESTIO_DEBUG
+		io_ptr = test_io_manager;
+		test_io_backing_manager = unix_io_manager;
+#else
+		io_ptr = unix_io_manager;
+#endif
 		retval = ext2fs_open(journal_device,
 				     EXT2_FLAG_JOURNAL_DEV_OK, 0,
-				     0, unix_io_manager, &jfs);
+				     0, io_ptr, &jfs);
 		if (retval) {
 			com_err(program_name, retval,
 				_("while trying to open journal device %s\n"),
 				journal_device);
 			exit(1);
 		}
+		if ((blocksize < 0) && (jfs->blocksize < -blocksize)) {
+			com_err(program_name, 0,
+				_("Journal dev blocksize (%d) smaller than "
+				  "minimum blocksize %d\n"), jfs->blocksize,
+				-blocksize);
+			exit(1);
+		}
 		blocksize = jfs->blocksize;
 		param.s_log_block_size =
 			int_log2(blocksize >> EXT2_MIN_BLOCK_LOG_SIZE);
@@ -1096,6 +1123,11 @@ static void PRS(int argc, char *argv[])
 				  "(max %d), forced to continue\n"),
 			blocksize, sys_page_size);
 	}
+	if ((blocksize > 4096) &&
+	    (param.s_feature_compat & EXT3_FEATURE_COMPAT_HAS_JOURNAL))
+		fprintf(stderr, "\nWarning: some 2.4 kernels do not support "
+			"blocksizes greater than 4096 \n\tusing ext3."
+			"  Use -b 4096 if this is an issue for you.\n\n");
 
 	if (param.s_feature_incompat & EXT3_FEATURE_INCOMPAT_JOURNAL_DEV) {
 		if (!fs_type)
@@ -1149,11 +1181,14 @@ static void PRS(int argc, char *argv[])
 				exit(1);
 			}
 			param.s_blocks_count = dev_size;
+			if (sys_page_size > EXT2_BLOCK_SIZE(&param))
+				param.s_blocks_count &= ~((sys_page_size /
+							   EXT2_BLOCK_SIZE(&param))-1);
 		}
 		
 	} else if (!force && (param.s_blocks_count > dev_size)) {
 		com_err(program_name, 0,
-			_("Filesystem larger than apparent filesystem size."));
+			_("Filesystem larger than apparent device size."));
 		proceed_question();
 	}
 
@@ -1171,11 +1206,20 @@ static void PRS(int argc, char *argv[])
 	    ((tmp = getenv("MKE2FS_FIRST_META_BG"))))
 		param.s_first_meta_bg = atoi(tmp);
 
-	set_fs_defaults(fs_type, &param, blocksize, &inode_ratio);
-
+	/* Get the hardware sector size, if available */
+	retval = ext2fs_get_device_sectsize(device_name, &sector_size);
+	if (retval) {
+		com_err(program_name, retval,
+			_("while trying to determine hardware sector size"));
+		exit(1);
+	}
+	
+	set_fs_defaults(fs_type, &param, blocksize, sector_size, &inode_ratio);
+	blocksize = EXT2_BLOCK_SIZE(&param);
+	
 	if (param.s_blocks_per_group) {
 		if (param.s_blocks_per_group < 256 ||
-		    param.s_blocks_per_group > group_blk_max) {
+		    param.s_blocks_per_group > 8 * blocksize) {
 			com_err(program_name, 0,
 				_("blocks per group count out of range"));
 			exit(1);
@@ -1189,7 +1233,7 @@ static void PRS(int argc, char *argv[])
 			com_err(program_name, 0,
 				_("bad inode size %d (min %d/max %d)"),
 				inode_size, EXT2_GOOD_OLD_INODE_SIZE,
-				EXT2_BLOCK_SIZE(&param));
+				blocksize);
 			exit(1);
 		}
 		if (inode_size != EXT2_GOOD_OLD_INODE_SIZE)
@@ -1203,7 +1247,7 @@ static void PRS(int argc, char *argv[])
 	 * Calculate number of inodes based on the inode ratio
 	 */
 	param.s_inodes_count = num_inodes ? num_inodes : 
-		((__u64) param.s_blocks_count * EXT2_BLOCK_SIZE(&param))
+		((__u64) param.s_blocks_count * blocksize)
 			/ inode_ratio;
 
 	/*
@@ -1220,6 +1264,7 @@ int main (int argc, char *argv[])
 	badblocks_list	bb_list = 0;
 	int		journal_blocks;
 	int		i, val;
+	io_manager	io_ptr;
 
 #ifdef ENABLE_NLS
 	setlocale(LC_MESSAGES, "");
@@ -1229,11 +1274,18 @@ int main (int argc, char *argv[])
 #endif
 	PRS(argc, argv);
 
+#ifdef CONFIG_TESTIO_DEBUG
+	io_ptr = test_io_manager;
+	test_io_backing_manager = unix_io_manager;
+#else
+	io_ptr = unix_io_manager;
+#endif
+
 	/*
 	 * Initialize the superblock....
 	 */
 	retval = ext2fs_initialize(device_name, 0, &param,
-				   unix_io_manager, &fs);
+				   io_ptr, &fs);
 	if (retval) {
 		com_err(device_name, retval, _("while setting up superblock"));
 		exit(1);
