@@ -48,7 +48,7 @@ extern int optind;
 #include "e2p/e2p.h"
 #include "jfs_user.h"
 #include "util.h"
-#include "get_device_by_label.h"
+#include "blkid/blkid.h"
 
 #include "../version.h"
 #include "nls-enable.h"
@@ -109,12 +109,13 @@ static void remove_journal_device(ext2_filsys fs)
 	int		i, nr_users;
 	errcode_t	retval;
 	int		commit_remove_journal = 0;
+	io_manager	io_ptr;
 
 	if (f_flag)
 		commit_remove_journal = 1; /* force removal even if error */
 
 	uuid_unparse(fs->super->s_journal_uuid, buf);
-	journal_path = get_spec_by_uuid(buf);
+	journal_path = blkid_get_devname(NULL, "UUID", buf);
 
 	if (!journal_path) {
 		journal_path =
@@ -123,9 +124,15 @@ static void remove_journal_device(ext2_filsys fs)
 			return;
 	}
 
+#ifdef CONFIG_TESTIO_DEBUG
+	io_ptr = test_io_manager;
+	test_io_backing_manager = unix_io_manager;
+#else
+	io_ptr = unix_io_manager;
+#endif
 	retval = ext2fs_open(journal_path, EXT2_FLAG_RW|
 			     EXT2_FLAG_JOURNAL_DEV_OK, 0,
-			     fs->blocksize, unix_io_manager, &jfs);
+			     fs->blocksize, io_ptr, &jfs);
 	if (retval) {
 		com_err(program_name, retval,
 			_("while trying to open external journal"));
@@ -184,8 +191,7 @@ no_valid_journal:
 		exit(1);
 	}
 	fs->super->s_journal_dev = 0;
-	memset(fs->super->s_journal_uuid, 0,
-	       sizeof(fs->super->s_journal_uuid));
+	uuid_clear(fs->super->s_journal_uuid);
 	ext2fs_mark_super_dirty(fs);
 	printf(_("Journal removed\n"));
 	free(journal_path);
@@ -359,6 +365,7 @@ static void add_journal(ext2_filsys fs)
 	unsigned long journal_blocks;
 	errcode_t	retval;
 	ext2_filsys	jfs;
+	io_manager	io_ptr;
 
 	if (fs->super->s_feature_compat &
 	    EXT3_FEATURE_COMPAT_HAS_JOURNAL) {
@@ -368,9 +375,15 @@ static void add_journal(ext2_filsys fs)
 	if (journal_device) {
 		check_plausibility(journal_device);
 		check_mount(journal_device, 0, _("journal"));
+#ifdef CONFIG_TESTIO_DEBUG
+		io_ptr = test_io_manager;
+		test_io_backing_manager = unix_io_manager;
+#else
+		io_ptr = unix_io_manager;
+#endif
 		retval = ext2fs_open(journal_device, EXT2_FLAG_RW|
 				     EXT2_FLAG_JOURNAL_DEV_OK, 0,
-				     fs->blocksize, unix_io_manager, &jfs);
+				     fs->blocksize, io_ptr, &jfs);
 		if (retval) {
 			com_err(program_name, retval,
 				_("\n\twhile trying to open journal on %s\n"),
@@ -427,7 +440,7 @@ static void parse_e2label_options(int argc, char ** argv)
 		fprintf(stderr, _("Usage: e2label device [newlabel]\n"));
 		exit(1);
 	}
-	device_name = argv[1];
+	device_name = blkid_get_devname(NULL, argv[1], NULL);
 	if (argc == 3) {
 		open_flag = EXT2_FLAG_RW | EXT2_FLAG_JOURNAL_DEV_OK;
 		L_flag = 1;
@@ -444,7 +457,18 @@ static time_t parse_time(char *str)
 		return (time(0));
 	}
 	memset(&ts, 0, sizeof(ts));
+#ifdef HAVE_STRPTIME
 	strptime(optarg, "%Y%m%d%H%M%S", &ts);
+#else
+	sscanf(optarg, "%4d%2d%2d%2d%2d%2d", &ts.tm_year, &ts.tm_mon,
+	       &ts.tm_mday, &ts.tm_hour, &ts.tm_min, &ts.tm_sec);
+	ts.tm_year -= 1900;
+	ts.tm_mon -= 1;
+	if (ts.tm_year < 0 || ts.tm_mon < 0 || ts.tm_mon > 11 ||
+	    ts.tm_mday < 0 || ts.tm_mday > 31 || ts.tm_hour > 23 ||
+	    ts.tm_min > 59 || ts.tm_sec > 61)
+		ts.tm_mday = 0;
+#endif
 	if (ts.tm_mday == 0) {
 		com_err(program_name, 0,
 			_("Couldn't parse date/time specifier: %s"),
@@ -666,7 +690,7 @@ static void parse_tune2fs_options(int argc, char **argv)
 		usage();
 	if (!open_flag && !l_flag)
 		usage();
-	device_name = argv[optind];
+	device_name = blkid_get_devname(NULL, argv[optind], NULL);
 }
 
 void do_findfs(int argc, char **argv)
@@ -678,7 +702,7 @@ void do_findfs(int argc, char **argv)
 		fprintf(stderr, "Usage: findfs LABEL=<label>|UUID=<uuid>\n");
 		exit(2);
 	}
-	dev = interpret_spec(argv[1]);
+	dev = blkid_get_devname(NULL, argv[1], NULL);
 	if (!dev) {
 		fprintf(stderr, "Filesystem matching %s not found\n", 
 			argv[1]);
@@ -694,6 +718,7 @@ int main (int argc, char ** argv)
 	errcode_t retval;
 	ext2_filsys fs;
 	struct ext2_super_block *sb;
+	io_manager io_ptr;
 
 #ifdef ENABLE_NLS
 	setlocale(LC_MESSAGES, "");
@@ -712,8 +737,13 @@ int main (int argc, char ** argv)
 	else
 		parse_tune2fs_options(argc, argv);
 	
-	retval = ext2fs_open (device_name, open_flag, 0, 0,
-			      unix_io_manager, &fs);
+#ifdef CONFIG_TESTIO_DEBUG
+	io_ptr = test_io_manager;
+	test_io_backing_manager = unix_io_manager;
+#else
+	io_ptr = unix_io_manager;
+#endif
+	retval = ext2fs_open (device_name, open_flag, 0, 0, io_ptr, &fs);
         if (retval) {
 		com_err (program_name, retval, _("while trying to open %s"),
 			 device_name);
